import java.util.Vector; 
import java.io.*;
import java.awt.print.*; 

/* Copyright K. Lano 2015-2019
   Package: QVT
*/ 

public class RelationalTransformation extends NamedElement
{ Vector modelParameter = new Vector(); // of TypedModel
  Vector rule = new Vector(); // of QVTRule
  Vector functions = new Vector(); // of BehaviouralFeature


  public RelationalTransformation(String n)
  { super(n); }

  public int nops() 
  { return functions.size(); } 

  public int nrules()
  { return rule.size(); } 

  public void addParameter(TypedModel tm)
  { modelParameter.add(tm); }

  public void addRule(QVTRule r)
  { rule.add(r); }

  public void addFunction(BehaviouralFeature bf)
  { functions.add(bf); }

  public boolean typeCheck(Vector types, Vector entities, Vector contexts, Vector env)
  { for (int i = 0; i < functions.size(); i++)
    { BehaviouralFeature bf = (BehaviouralFeature) functions.get(i);
      bf.typeCheck(types,entities); 
    }

    for (int i = 0; i < rule.size(); i++)
    { QVTRule r = (QVTRule) rule.get(i);
      r.typeCheck(types,entities,contexts,env); 
    } 
    return true; 
  } 

  public void addTraceEntities(Vector entities, Vector assocs)
  { for (int i = 0; i < rule.size(); i++)
    { QVTRule r = (QVTRule) rule.get(i);
      r.addTraceEntity(entities,assocs); 
    } 
  } 

  public Vector toUseCase(Vector entities, Vector types)
  { String nme = getName(); 
    UseCase pres = new UseCase(nme + "$Pres", null); 
    UseCase res = new UseCase(nme + "$Con",null);
    UseCase cleanup = new UseCase(nme + "$Cleanup",null);

    Vector ucs = new Vector(); 
    ucs.add(pres); 
    ucs.add(res); 
    ucs.add(cleanup); 

    // The functions are added to pres and res
 
    for (int i = 0; i < functions.size(); i++)
    { BehaviouralFeature bf = (BehaviouralFeature) functions.get(i);
      res.addOperation(bf);  
    } 

    for (int i = 0; i < rule.size(); i++)
    { Relation r = (Relation) rule.get(i);
      r.getNontopOp(pres,res,entities,types); 
      // Vector bnd = new Vector(); 
      // Vector guard = r.toGuardCondition(bnd,entities); 
      // System.out.println("*** Guard condition of " + r.getName() + " is " + guard); 
    } // only needed if there is a where that calls it

    Vector ruleclones = new Vector(); 

    for (int i = 0; i < rule.size(); i++)
    { Relation r = (Relation) rule.get(i);
      Relation r1 = (Relation) r.clone(); 
      ruleclones.add(r1); 
    } 

    for (int i = 0; i < rule.size(); i++)
    { QVTRule r = (QVTRule) rule.get(i);
      // Constraint c = r.toInverseConstraint(entities,types);
      // if (c != null)
      // { res.addPostcondition(c); }
      pres.addPostconditions(r.toPreservationConstraints(res,entities,types)); 
    }
      
    for (int i = 0; i < ruleclones.size(); i++)
    { QVTRule r1 = (QVTRule) ruleclones.get(i); 
      Vector c = r1.toConstraints(entities,types);
      if (c != null)
      { res.addPostconditions(c); }
    }

    Vector tents = new Vector(); 
    for (int i = 0; i < rule.size(); i++) 
    { QVTRule r = (QVTRule) rule.get(i); 
      tents = VectorUtil.union(tents, r.targetEntities()); 
    } 

    for (int i = 0; i < tents.size(); i++)
    { Entity e = (Entity) tents.get(i);
      Constraint cc = toCleanupConstraint(e,entities); 
      if (cc != null) 
      { cleanup.addPostcondition(cc); }  
    }
    
    return ucs;
  }

  public Constraint toCleanupConstraint(Entity e, Vector entities)
  { // not(R$trace->exists( tr ! tr.var = self )) and ... and not(...) => self->isDeleted()
    // for each R with target attribute  var : e

    String ename = e.getName(); 
    Expression ante = new BasicExpression(true); 
    Vector allArgs = new Vector(); 

    for (int i = 0; i < rule.size(); i++) 
    { QVTRule r = (QVTRule) rule.get(i); 

      if (r.isAbstract()) { continue; } 

      Entity trent = (Entity) ModelElement.lookupByName(r.getName() + "$trace", entities); 
      if (trent == null) 
      { return null; } 

      Vector targs = r.targetAttributes(); 
      for (int j = 0; j < targs.size(); j++) 
      { Attribute tatt = (Attribute) targs.get(j); 
        if ((tatt.getType() + "").equals(ename))
        { Expression checktr = r.checkTraceEmpty(trent,tatt); 
          // UnaryExpression anter = new UnaryExpression("not", checktr); 
          ante = Expression.simplifyAnd(ante,checktr); 
          allArgs.add(tatt); 
        } 
      } 
    } 

    if (allArgs.size() == 0) { return null; } 

    Attribute att = (Attribute) allArgs.get(0); 
        
    Expression delrd = Relation.deleteDomainExp(att); 
    Constraint cc = new Constraint(ante,delrd);
    cc.setOwner(e); 
    return cc; 
  }
       
  public String toString()
  { String res = "transformation " + getName() + "\n" + 
      "{ "; 
    
    for (int i = 0; i < functions.size(); i++) 
    { BehaviouralFeature f = (BehaviouralFeature) functions.get(i);      
      res = res + "query " + f.getSignature() + " : " + f.getResultType() + "\n" + "{ " + f.getPost() + " }\n\n"; 
    }

    for (int i = 0; i < rule.size(); i++) 
    { QVTRule r = (QVTRule) rule.get(i);      
      res = res + r + "\n\n"; 
    } 
    return res + "}"; 
  } 

  public int complexity(PrintWriter out)
  { int res = 0; 
    for (int i = 0; i < functions.size(); i++) 
    { BehaviouralFeature f = (BehaviouralFeature) functions.get(i);      
      int fc = f.syntacticComplexity(); 
      out.println("**** operation " + f.getName() + " has complexity " + fc); 
      if (fc > 100) 
      { out.println("**** EHS flaw, excessively large helper"); }  
      res = res + fc; 
    } 

    for (int i = 0; i < rule.size(); i++) 
    { QVTRule r = (QVTRule) rule.get(i);      
      int rc = r.complexity(); 
      out.println("**** rule " + r.getName() + " has complexity " + rc); 
      if (rc > 100) 
      { out.println("**** ERS flaw, excessively large rule"); }  

      res = res + rc; 
    } 

    return res; 
  } 

  public int cyclomaticComplexity()
  { int res = 0; 
    for (int i = 0; i < functions.size(); i++) 
    { BehaviouralFeature f = (BehaviouralFeature) functions.get(i);      
      if (f.cyclomaticComplexity() > 10) { res++; }  
    } 

    for (int i = 0; i < rule.size(); i++) 
    { QVTRule r = (QVTRule) rule.get(i);      
      if (r.cyclomaticComplexity() > 10) { res++; }  
    } 

    return res; 
  } 

  public int enr() { return rule.size(); } 

  public int eno() { return functions.size(); } 

  public int uex()
  { int res = 0; 
    for (int i = 0; i < rule.size(); i++) 
    { Relation r = (Relation) rule.get(i);      
      if (r.isTopLevel) 
      { res++; }  
    } 
    return (res*(res-1))/2; 
  } 

  public int epl(PrintWriter out)
  { int res = 0; 
    for (int i = 0; i < rule.size(); i++) 
    { Relation r = (Relation) rule.get(i);      
      int p = r.epl(); 
      if (p > 10) 
      { out.println("*** EPL flaw: " + p + " parameters for rule " + r.getName()); 
        res++; 
      }  
    } 
    return res; 
  }

  public Map getCallGraph()
  { Map res = new Map(); 
    int efo = 0; 

    // operations, pre and post also
    for (int i = 0; i < functions.size(); i++) 
    { BehaviouralFeature rr = (BehaviouralFeature) functions.get(i); 
      rr.getCallGraph(res); 
    } 

    for (int i = 0; i < rule.size(); i++) 
    { Relation r = (Relation) rule.get(i);      
      r.getCallGraph(res);  
    } 

    return res; 
  }  

}
